#![feature(test, platform_intrinsics)]

extern crate simd;
extern crate test;

use simd::u8x16;

const N: usize = 1000000;

#[bench]
fn index(b: &mut test::Bencher) {
    let rgb = vec![0_u8; N * 3];
    let mut rgba = vec![0; N * 4];

    b.iter(|| {
        for i in 0..N {
            rgba[4 * i + 0] = rgb[3 * i + 0];
            rgba[4 * i + 1] = rgb[3 * i + 1];
            rgba[4 * i + 2] = rgb[3 * i + 2];
            rgba[4 * i + 3] = 255;
        }
        test::black_box(&rgba);
    })
}

#[bench]
fn manual_iter(b: &mut test::Bencher) {
    let rgb = vec![0_u8; N * 3];
    let mut rgba = vec![0; N * 4];

    b.iter(|| {
        {
            let mut ait = rgb.iter();
            let mut bit = rgba.iter_mut();
            while let (Some(&r),Some(&g),Some(&b)) = (ait.next(),ait.next(),ait.next()) {
                if let (Some(rr),Some(gg),Some(bb),Some(aa)) = (bit.next(),bit.next(),bit.next(),bit.next()) {
                    *rr = b;
                    *gg = g;
                    *bb = r;
                    *aa = 255;
                } else {
                    panic!("out of bounds");
                }
            }
        }
        test::black_box(&rgba);
    })
}

#[bench]
fn unsafe_get(b: &mut test::Bencher) {
    let rgb = vec![0_u8; N * 3];
    let mut rgba = vec![0; N * 4];

    b.iter(|| {
        for i in 0..N {
            unsafe {
                *rgba.get_unchecked_mut(4 * i + 0) = *rgb.get_unchecked(3 * i + 0);
                *rgba.get_unchecked_mut(4 * i + 1) = *rgb.get_unchecked(3 * i + 1);
                *rgba.get_unchecked_mut(4 * i + 2) = *rgb.get_unchecked(3 * i + 2);
                *rgba.get_unchecked_mut(4 * i + 3) = 255;
            }
        }
        test::black_box(&rgba);
    })
}


fn rgba_conv(x: &[u8], y: &mut [u8]) {
    assert!(x.len() % 3 * 16 == 0);
    assert!(y.len() % 4 * 16 == 0);
    for i in 0..x.len() / (3 * 16) {
        let xi = i * 3 * 16;
        let a = u8x16::load(x, xi + 16 * 0);
        let b = u8x16::load(x, xi + 16 * 1);
        let c = u8x16::load(x, xi + 16 * 2);

        let p = u8x16::new(a.extract(3 * 0 + 0), a.extract(3 * 0 + 1), a.extract(3 * 0 + 2), 255,
                           a.extract(3 * 1 + 0), a.extract(3 * 1 + 1), a.extract(3 * 1 + 2), 255,
                           a.extract(3 * 2 + 0), a.extract(3 * 2 + 1), a.extract(3 * 2 + 2), 255,
                           a.extract(3 * 3 + 0), a.extract(3 * 3 + 1), a.extract(3 * 3 + 2), 255);
        let q = u8x16::new(a.extract(3 * 4 + 0), a.extract(3 * 4 + 1), a.extract(3 * 4 + 2), 255,
                           a.extract(3 * 5 + 0), b.extract(3 * 0 + 0), b.extract(3 * 0 + 1), 255,
                           b.extract(3 * 0 + 2), b.extract(3 * 1 + 0), b.extract(3 * 1 + 1), 255,
                           b.extract(3 * 1 + 2), b.extract(3 * 2 + 0), b.extract(3 * 2 + 1), 255);
        let r = u8x16::new(b.extract(3 * 2 + 2), b.extract(3 * 3 + 0), b.extract(3 * 3 + 1), 255,
                           b.extract(3 * 3 + 2), b.extract(3 * 4 + 0), b.extract(3 * 4 + 1), 255,
                           b.extract(3 * 4 + 2), b.extract(3 * 5 + 0), c.extract(3 * 0 + 0), 255,
                           c.extract(3 * 0 + 1), c.extract(3 * 0 + 2), c.extract(3 * 1 + 0), 255);
        let s = u8x16::new(c.extract(3 * 1 + 1), c.extract(3 * 1 + 2), c.extract(3 * 2 + 0), 255,
                           c.extract(3 * 2 + 1), c.extract(3 * 2 + 2), c.extract(3 * 3 + 0), 255,
                           c.extract(3 * 3 + 1), c.extract(3 * 3 + 2), c.extract(3 * 4 + 0), 255,
                           c.extract(3 * 4 + 1), c.extract(3 * 4 + 2), c.extract(3 * 5 + 0), 255);
        let yi = i * 4 * 16;
        p.store(y, yi + 16 * 0);
        q.store(y, yi + 16 * 1);
        r.store(y, yi + 16 * 2);
        s.store(y, yi + 16 * 3);
    }
}

#[bench]
fn simd_(b: &mut test::Bencher) {
    let rgb = vec![0_u8; N * 3];
    let mut rgba = vec![0; N * 4];

    b.iter(|| {
        rgba_conv(&rgb, &mut rgba);
        test::black_box(&rgba);
    })
}

fn rgba_conv_shuf(x: &[u8], y: &mut [u8]) {
    assert!(x.len() % 3 * 16 == 0);
    assert!(y.len() % 4 * 16 == 0);
    for i in 0..x.len() / (3 * 16) {
        let xi = i * 3 * 16;
        let a = u8x16::load(x, xi + 16 * 0);
        let b = u8x16::load(x, xi + 16 * 1);
        let c = u8x16::load(x, xi + 16 * 2);

        extern"platform-intrinsic" {
            fn simd_shuffle16<T, U>(x: T, y: T, idx: [u32; 16]) -> U;
        }
        let _255 = u8x16::splat(255);
        unsafe {
            let p: u8x16 = simd_shuffle16(a, _255, [3 * 0 + 0 , 3 * 0 + 1, 3 * 0 + 2, 16,
                                                    3 * 1 + 0 , 3 * 1 + 1, 3 * 1 + 2, 16,
                                                    3 * 2 + 0 , 3 * 2 + 1, 3 * 2 + 2, 16,
                                                    3 * 3 + 0 , 3 * 3 + 1, 3 * 3 + 2, 16]);

            let mut q: u8x16 = simd_shuffle16(a, b, [3 * 4 + 0, 3 * 4 + 1, 3 * 4 + 2, 3 * 5 + 0,
                                                     16 + 3 * 0 + 0, 16 + 3 * 0 + 1, 16 + 3 * 0 + 2, 16 + 3 * 1 + 0,
                                                     16 + 3 * 1 + 1, 16 + 3 * 1 + 2, 16 + 3 * 2 + 0, 16 + 3 * 2 + 1,
                                                     0, 0, 0, 0]);
            q = simd_shuffle16(q, _255, [3 * 0 + 0 , 3 * 0 + 1, 3 * 0 + 2, 16,
                                         3 * 1 + 0 , 3 * 1 + 1, 3 * 1 + 2, 16,
                                         3 * 2 + 0 , 3 * 2 + 1, 3 * 2 + 2, 16,
                                         3 * 3 + 0 , 3 * 3 + 1, 3 * 3 + 2, 16]);
            let mut r: u8x16 = simd_shuffle16(b, c, [3 * 2 + 2, 3 * 3 + 0, 3 * 3 + 1, 3 * 3 + 2,
                                                     3 * 4 + 0, 3 * 4 + 1, 3 * 3 + 2, 3 * 5 + 0,
                                                     16 + 3 * 0 + 0, 16 + 3 * 0 + 1, 16 + 3 * 0 + 2, 16 + 3 * 1 + 0,
                                                     0, 0, 0, 0]);
            r = simd_shuffle16(r, _255, [3 * 0 + 0 , 3 * 0 + 1, 3 * 0 + 2, 16,
                                         3 * 1 + 0 , 3 * 1 + 1, 3 * 1 + 2, 16,
                                         3 * 2 + 0 , 3 * 2 + 1, 3 * 2 + 2, 16,
                                         3 * 3 + 0 , 3 * 3 + 1, 3 * 3 + 2, 16]);

            let s: u8x16 = simd_shuffle16(c, _255, [3 * 1 + 1, 3 * 1 + 2, 3 * 2 + 0, 16,
                                                    3 * 2 + 1, 3 * 2 + 2, 3 * 3 + 0, 16,
                                                    3 * 3 + 1, 3 * 3 + 2, 3 * 4 + 0, 16,
                                                    3 * 4 + 1, 3 * 4 + 2, 3 * 5 + 0, 16]);
            let yi = i * 4 * 16;
            p.store(y, yi + 16 * 0);
            q.store(y, yi + 16 * 1);
            r.store(y, yi + 16 * 2);
            s.store(y, yi + 16 * 3);
        }
    }
}

#[bench]
fn simd_explicit_shuffle(b: &mut test::Bencher) {
    let rgb = vec![0_u8; N * 3];
    let mut rgba = vec![0; N * 4];

    b.iter(|| {
        rgba_conv_shuf(&rgb, &mut rgba);
        test::black_box(&rgba);
    })
}
